"""
This module contains LCSD class which holds all necessary parameters for logarithmic spectral density (LCSD/LPSD)
calculation
"""
from typing import Callable, Optional, Union
from warnings import warn

import numpy as np
from pandas import DataFrame, Series
from pandas.core.indexes.datetimes import DatetimeIndex

from ._helpers import (
    _calc_lcsd,
    _calc_lcsd_py,
    _kaiser_alpha,
    _kaiser_rov,
    _ltf_plan,
    c_core_available,
)


class LCSD:  # pylint: disable=too-many-instance-attributes
    """
    Class to hold paeameters for LCSD/LPSD calculation
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        data: Union[Series, DataFrame],
        sample_rate: Optional[
            float
        ] = None,  # None: Use time series data from first column to calculate sampling rate
        window_function: Callable = np.kaiser,
        overlap: Optional[float] = None,  # None: use default overlap
        detrending_order: Optional[int] = 0,
        n_frequencies: int = 1000,  # also called Jdes
        n_averages: int = 100,  # known as Kdes
        n_min_bins: int = 1,  # bmin
        min_segment_length: float = 0,  # known as Lmin
        psll: float = 200,
        use_c_core: bool = True,
        csd: bool = False,
    ):
        """
        Parameters
        ----------
        data: :obj:`pandas.Series` or :obj:`pandas.DataFrame`
            Assuming time series data and takes one column of a :obj:`pandas.DataFrame`
            or a whole `DataFrame`. If multiple columns are provided, it will calculate
            the spectrum for each column and return a dict of `DataFrame`s with all
            results. Unless csd=True, in which case input DataFrame is assumed to have
            exactly two columns with two input series.
        sample_rate: :obj:`float` (optional)
            Sampling rate of the data. Defaults to calculating the mean difference of the first columns elements.
        window_function: :obj:`Callable` (optional)
            Define a window function, defaults to :obj:`np.kaiser`.
        overlap: :obj:`float` (optional)
            Overlap percentage, usually taken from the window function.
            Defaults to recommended overlap.
        detrending_order: :obj:`int` (optional)
            Order for detrending, 0 = offset, 1 = linear, :obj:`None` to disable.
            Also just called *order*.
        n_frequencies: :obj:`int` (optional)
            The desired number of frequencies.
            Also called *Jdes*.
        n_averages: :obj:`int` (optional)
            The desired number of averages
            Also called *Kdes*.
        n_min_bins: :obj:`int` (optional)
            The minimum bin number to be used, usually taken from the window function.
            Also called *bmin*.
        min_segment_length: :obj:`float` (optional)
            The minimum segment length.
            Also called *Lmin*.
        psll: :obj:`float` (optional)
            Peak side-lobe level.
        use_c_core: :obj:`bool`
            Use the C core or the pure Python implementation.
        csd: :obj:`bool`
            Calculate CSD instead of PSD: in this case, DataFrame is assumed to have two columns with two input series
        """
        self.data = data
        self.sample_rate = sample_rate
        self.window_function = window_function
        self.overlap = overlap
        self.detrending_order = detrending_order
        self.n_frequencies = n_frequencies
        self.n_averages = n_averages
        self.n_min_bins = n_min_bins
        self.min_segment_length = min_segment_length
        self.psll = psll
        self.use_c_core = use_c_core
        self.csd = csd
        # parameters for LFT planning:
        self._f: list = []
        self._r: list = []
        self._m: list = []
        self._L: list = []
        self._K: list = []

    def _run_lcsd(self, data1, data2):
        lcsd_runner = _calc_lcsd if self.use_c_core else _calc_lcsd_py
        result = lcsd_runner(
            data1.to_numpy(),
            data2.to_numpy(),
            self._f,
            self._r,
            self._m,
            self._L,
            self.sample_rate,
            self.window_function,
            self.psll,
            self.detrending_order,
            self.overlap,
            self.min_segment_length,
            self.csd,
        )

        dc = DataFrame(
            dict(
                ps=result[0],
                psd=result[1],
                ps_std=result[2],
                psd_std=result[3],
                enbw=result[4],
                asd=result[5],
                asdrms=result[6],
            ),
            index=self._f,
        )
        dc.index.name = "frequency"
        return dc

    def check_inputs(self):
        if self.sample_rate is None:
            if isinstance(self.data.index, DatetimeIndex):
                index_diff = self.data.index.to_series().diff()
                period_time = index_diff.median().total_seconds()
                std = index_diff.std().total_seconds()
            else:
                index_diff = np.diff(self.data.index)
                period_time = np.median(index_diff)
                std = index_diff.std()
            self.sample_rate = 1 / period_time

            if std / period_time > 1e-6:
                warn(
                    "Length of some time steps deviates a lot from the median. Some data maybe corrupt!\n"
                    f"Period time: {period_time}, standard deviation: {std}",
                    UserWarning,
                )

        if self.data.isna().any().any():
            warn("Removing NaN values ...", UserWarning)
            self.data.dropna(inplace=True)
        assert (
            not self.data.isna().any().any()
        ), "Bug: There should not be any NaN data!"

        if self.use_c_core and not c_core_available():
            warn(
                "C core should be used, but is not available. Using Python …",
                RuntimeWarning,
            )
            warn(
                f"use_c_core: {self.use_c_core}, c_core_available: {c_core_available()}",
                RuntimeWarning,
            )
            self.use_c_core = False

        if isinstance(self.data, DataFrame) and len(self.data.columns) == 1:
            self.data = self.data[self.data.columns[0]]

        if self.detrending_order is None:
            self.detrending_order = -1

        if self.csd:
            if not isinstance(self.data, DataFrame):
                raise ValueError(
                    "Only 2-column DataFrames are accepted as input for CSD"
                )
            if len(self.data.columns) != 2:
                raise ValueError(
                    "Only 2-column DataFrames are accepted as input for CSD"
                )
            if len(self.data[self.data.columns[0]].to_numpy()) != len(
                self.data[self.data.columns[1]].to_numpy()
            ):
                raise ValueError(
                    "Input time series have different lengths (number of samples)! "
                    "Only time series with equal lengths can be used to calculate CSD"
                )

    def run(self):
        if self.window_function is np.kaiser:
            # calculate kaiser parameters
            alpha = _kaiser_alpha(self.psll)
            if self.overlap is None:
                self.overlap = _kaiser_rov(alpha)

        (
            self._f,
            self._r,
            self._m,
            self._L,
            self._K,
        ) = _ltf_plan(  # pylint: disable=unused-variable
            len(self.data),
            self.sample_rate,
            self.overlap,
            self.n_min_bins,
            self.min_segment_length,
            self.n_frequencies,
            self.n_averages,
        )

        if self.csd:
            dc = self._run_lcsd(
                self.data[self.data.columns[0]], self.data[self.data.columns[1]]
            )
        elif isinstance(self.data, DataFrame):
            dc = {}
            for col in self.data.columns:
                dc[col] = self._run_lcsd(self.data[col], self.data[col])
        else:
            dc = self._run_lcsd(self.data, self.data)

        return dc



"""This module implements several flat-top window functions.

It is based on a Matlab script from Joe Henning from 6 May 2013.

Implemented flat-top window functions:
SFT3F, 31.7 dB PSLL, differentiable, 0.0082 dB emax
SFT4F, 44.7 dB PSLL, 2nd differentiable, 0.0041 dB emax
SFT5F, 57.3 dB PSLL, 3rd differentiable, -0.0025 dB emax
SFT3M, 44.2 dB PSLL, -0.0115 dB emax
SFT4M, 66.5 dB PSLL, -0.0067 dB emax
SFT5M, 89.9 dB PSLL, 0.0039 dB emax
FTNI (National Instruments), 44.4 dB PSLL, 0.0169 dB emax
FTHP (old Hewlett Packard), 70.4 dB PSLL, 0.0096 dB emax
FTSRS (Stanford Research SR785), 76.6 dB PSLL, differentiable, -0.0156 dB emax
Matlab, 93.0 dB PSLL, 0.0097 dB emax
HFT70 (3-term cosine), 70.4 dB PSLL, -0.0065 dB emax
HFT95 (4-term cosine), 95.0 dB PSLL, 0.0044 dB emax
HFT90D (4-term cosine), 90.2 dB PSLL, differentiable, -0.0039 dB emax
HFT116D (5-term cosine), 116.8 dB PSLL, differentiable, -0.0028 dB emax
HFT144D (6-term cosine), 144.1 dB PSLL, differentiable, 0.0021 dB emax
HFT169D (7-term cosine), 169.5 dB PSLL, differentiable, 0.0017 dB emax
HFT196D (8-term cosine), 196.2 dB PSLL, differentiable, 0.0013 dB emax
HFT223D (9-term cosine), 223.0 dB PSLL, differentiable, -0.0011 dB emax
HFT248D (10-term cosine), 248.4 dB PSLL, differentiable, 0.0009 dB emax

In the above, "F" indicates "fast decaying" which enforces a high
degree of differentiability, "M" indicates minimum sidelobe, and
"HFT" indicates the window with lowest sidelobe level.

It also includes a dictionary in which the optimal overlap values are given.

Ref:
Spectrum and spectral density estimation by the Discrete Fourier transform (DFT), including a comprehensive list of window functions and some new flat-top windows
G. Heinzel, A. Rudiger, and R. Schilling
Maz-Planck-Institut fur Gravitationsphysik
(Albert-Einstein-Institut)
Teilinstitut Hannover
February 15, 2002
"""
import numpy as np


def SFT3F(M):
    """Flat-top window function SFT3F.

    Differentiable, -31.7 dB, NBW 3.1681 bins, 0.0082 dB, first zero at +/- 3 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.26526
    a1 = 0.5
    a2 = 0.23474
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z)
    return w


def SFT4F(M):
    """Flat-top window function SFT4F.

    2nd differentiable, -44.7 dB, NBW 3.7970 bins, 0.0041 dB, first zero at +/- 4 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.21706
    a1 = 0.42103
    a2 = 0.28294
    a3 = 0.07897
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z) - a3 * np.cos(6 * z)
    return w


def SFT5F(M):
    """Flat-top window function SFT5F.

    3rd differentiable, -57.3 dB, NBW 4.3412 bins, -0.0025 dB, first zero at +/- 5 bins

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.1881
    a1 = 0.36923
    a2 = 0.28702
    a3 = 0.13077
    a4 = 0.02488
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def SFT3M(M):
    """Flat-top window function SFT3M.

    -44.2 dB, NBW 2.9452 bins, -0.0115 dB, first zero at +/- 3 bins

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.28235
    a1 = 0.52105
    a2 = 0.19659
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z)
    return w


def SFT4M(M):
    """Flat-top window function SFT4M.

    -66.5 dB, NBW 3.3868 bins, -0.0067 dB, first zero at +/- 4 bins

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.241906
    a1 = 0.460841
    a2 = 0.255381
    a3 = 0.041872
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z) - a3 * np.cos(6 * z)
    return w


def SFT5M(M):
    """Flat-top window function SFT5F.

    -89.9 dB, NBW 3.8852 bins, 0.0039 dB, first zero at +/- 5 bins

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.209671
    a1 = 0.407331
    a2 = 0.281225
    a3 = 0.092669
    a4 = 0.0091036
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def FTNI(M):
    """Flat-top window function FTNI (National Instruments).

    -44.4 dB, NBW 2.9656 bins, 0.0169 dB, first zero at +/- 3 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.2810639
    a1 = 0.5208972
    a2 = 0.1980399
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z)
    return w


def FTHP(M):
    """Flat-top window function # FTHP (old Hewlett Packard).

    -70.4 dB, NBW 3.4279 bins, 0.0096 dB, first zero at +/- 4 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.912510941
    a2 = 1.079173272
    a3 = 0.1832630879
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z) - a3 * np.cos(6 * z)
    return w


def FTSR(M):
    """Flat-top window function FTSRS (Stanford Research SR785).

    Differentiable, -76.6 dB, NBW 3.7702 bins, -0.0156 dB, first zero at +/- 4.72 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.93
    a2 = 1.29
    a3 = 0.388
    a4 = 0.028
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def Matlab(M):
    """Flat-top window function Matlab.

    -93.0 dB, NBW 3.774 bins, 0.0025 dB, first zero at +/- 5.01 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 0.21557895
    a1 = 0.41663158
    a2 = 0.277263158
    a3 = 0.083578947
    a4 = 0.006947368
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def HFT70(M):
    """Flat-top window function HFT70 (lowest sidelobe level with 3 np.cosine terms).

    -70.4 dB, NBW 3.4129 bins, -0.0065 dB, first zero at +/- 4 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """

    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.90796
    a2 = 1.07349
    a3 = 0.18199
    w = a0 - a1 * np.cos(2 * z) + a2 * np.cos(4 * z) - a3 * np.cos(6 * z)
    return w


def HFT95(M):
    """Flat-top window function HFT95 (lowest sidelobe level with 4 np.cosine terms),

    -95.0 dB, NBW 3.8112 bins, 0.0044 dB, first zero at +/- 5 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.9383379
    a2 = 1.3045202
    a3 = 0.4028270
    a4 = 0.0350665
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def HFT90D(M):
    """Flat-top window function HFT90D (lowest sidelobe level with 4 np.cosine terms).

    Differentiable, -90.2, NBW 3.8832 bins, -0.0039 dB, first zero at +/- 5 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.942604
    a2 = 1.340318
    a3 = 0.440811
    a4 = 0.043097
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
    )
    return w


def HFT116D(M):
    """Flat-top window function HFT116D (lowest sidelobe level with 5 np.cosine terms).

    Differentiable, -116.8 dB, NBW 4.2186 bins, -0.0028 dB, first zero at +/- 6 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.9575375
    a2 = 1.4780705
    a3 = 0.6367431
    a4 = 0.1228389
    a5 = 0.0066288
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
    )
    return w


def HFT144D(M):
    """Flat-top window function HFT144D (lowest sidelobe level with 6 np.cosine terms).

    Differentiable, -144.1 dB, NBW 4.5386 bins, 0.0021 dB, first zero at +/- 7 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.96760033
    a2 = 1.57983607
    a3 = 0.81123644
    a4 = 0.22583558
    a5 = 0.02773848
    a6 = 0.00090360
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
        + a6 * np.cos(12 * z)
    )
    return w


def HFT169D(M):
    """Flat-top window function HFT169D (lowest sidelobe level with 7 np.cosine terms).

    Differentiable, -169.5 dB, NBW 4.8347 bins, 0.0017 dB, first zero at +/- 8 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.97441842
    a2 = 1.65409888
    a3 = 0.95788186
    a4 = 0.33673420
    a5 = 0.06364621
    a6 = 0.00521942
    a7 = 0.00010599
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
        + a6 * np.cos(12 * z)
        - a7 * np.cos(14 * z)
    )
    return w


def HFT196D(M):
    """Flat-top window function HFT196D (lowest sidelobe level with 8 np.cosine terms).

    Differentiable, -196.2 dB, NBW 5.1134 bins, 0.0013 dB, first zero at +/- 9 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.979280420
    a2 = 1.710288951
    a3 = 1.081629853
    a4 = 0.448734314
    a5 = 0.112376628
    a6 = 0.015122992
    a7 = 0.000871252
    a8 = 0.000011896
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
        + a6 * np.cos(12 * z)
        - a7 * np.cos(14 * z)
        + a8 * np.cos(16 * z)
    )
    return w


def HFT223D(M):
    """Flat-top window function HFT223D (lowest sidelobe level with 9 np.cosine terms).

    Differentiable, -223.0 dB, NBW 5.3888 bins, -0.0011 dB, first zero at +/- 10 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.98298997309
    a2 = 1.75556083063
    a3 = 1.19037717712
    a4 = 0.56155440797
    a5 = 0.17296769663
    a6 = 0.03233247087
    a7 = 0.00324954578
    a8 = 0.00013801040
    a9 = 0.00000132725
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
        + a6 * np.cos(12 * z)
        - a7 * np.cos(14 * z)
        + a8 * np.cos(16 * z)
        - a9 * np.cos(18 * z)
    )
    return w


def HFT248D(M):
    """Flat-top window function HFT248D (lowest sidelobe level with 10 np.cosine terms).

    Differentiable, -248.4 dB, NBW 5.6512 bins, 0.0009 dB, first zero at +/- 11 bins.

    Args:
        M (int): Points in the output window.

    Returns:
        np.array(): Window values as vector.
    """
    z = np.pi / M * np.arange(0, M)
    a0 = 1.0
    a1 = 1.985844164102
    a2 = 1.791176438506
    a3 = 1.282075284005
    a4 = 0.667777530266
    a5 = 0.240160796576
    a6 = 0.056656381764
    a7 = 0.008134974479
    a8 = 0.000624544650
    a9 = 0.000019808998
    a10 = 0.000000132974
    w = (
        a0
        - a1 * np.cos(2 * z)
        + a2 * np.cos(4 * z)
        - a3 * np.cos(6 * z)
        + a4 * np.cos(8 * z)
        - a5 * np.cos(10 * z)
        + a6 * np.cos(12 * z)
        - a7 * np.cos(14 * z)
        + a8 * np.cos(16 * z)
        - a9 * np.cos(18 * z)
        + a10 * np.cos(20 * z)
    )
    return w


olap_dict = {
    "Rectangular": 0.0,
    "Welch": 0.293,
    "Barlett": 0.500,
    "Hanning": 0.500,
    "Hamming": 0.500,
    "Nuttall3": 0.647,
    "Nuttall4": 0.705,
    "Nuttall3a": 0.612,
    "Kaiser3": 0.619,
    "Nuttall3b": 0.598,
    "Nuttall4a": 0.680,
    "BH92": 0.661,
    "Nuttall4b": 0.663,
    "Kaiser4": 0.670,
    "Nuttall4c": 0.656,
    "Kaiser5": 0.705,
    "SFT3F": 0.667,
    "SFT3M": 0.655,
    "FTNI": 0.656,
    "SFT4F": 0.750,
    "SFT5F": 0.785,
    "SFT4M": 0.721,
    "FTHP": 0.723,
    "HFT70": 0.722,
    "FTSRS": 0.754,
    "SFT5M": 0.760,
    "HFT90D": 0.760,
    "HFT95": 0.756,
    "HFT116D": 0.782,
    "HFT144D": 0.799,
    "HFT169D": 0.812,
    "HFT196D": 0.823,
    "HFT223D": 0.833,
    "HFT248D": 0.841,
}



import logging as log
import sys

from .wrapper import lpsd, lcsd, lpsd_trad

if sys.version_info >= (3, 8):
    from importlib import metadata  # >= Python 3.8
else:
    import importlib_metadata as metadata  # <= Python 3.7 pylint: disable=import-error

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    __version__ = "unknown"
    log.warning("Version not known, importlib.metadata is not working correctly.")



"""
This module implements the LPSD algorithm by G. Heinzel and M. Troebs.
For the core part of the algorithm it uses an C implementation by M. Hewitson/G.Heinzel.
Additionally a pure python implementation is provided.
It depends on the packages numpy and ctypes.
Ref: Improved spectrum estimation from digitized time series
on a logarithmic frequency axis
https://doi.org/10.1016/j.measurement.2005.10.010
"""
from typing import Union
from pandas import DataFrame

from ._lcsd import LCSD


def lcsd(data: DataFrame, *args, **kwargs) -> DataFrame:
    """
    Wrapper for LCSD assuming data in a :obj:`pandas.DataFrame`.
    Only 2-column DataFrames are accepted on input, to calculate CSD(col1,col2)
    The sample frequency will be calculated from the index column.

    Example
    -------
    .. code-block:: python

        from lpsd import lcsd
        result = lcsd(data) #  data is pandas.DataFrame with two columns

    Parameters
    ----------
    data: :obj:`pandas.DataFrame`
        Assuming time series data and takes one column of a :obj:`pandas.DataFrame`
        as signal1 and second column as signal2, returns `DataFrame`s with CSD(signal1,signal2)
        and all additional results
    sample_rate: :obj:`float` (optional)
        Sampling rate of the data. Defaults to calculating the mean difference of the first columns elements.
    window_function: :obj:`Callable` (optional)
        Define a window function, defaults to :obj:`np.kaiser`.
    overlap: :obj:`float` (optional)
        Overlap percentage, usually taken from the window function.
        Defaults to recommended overlap.
    detrending_order: :obj:`int` (optional)
        Order for detrending, 0 = offset, 1 = linear, :obj:`None` to disable.
        Also just called *order*.
    n_frequencies: :obj:`int` (optional)
        The desired number of frequencies.
        Also called *Jdes*.
    n_averages: :obj:`int` (optional)
        The desired number of averages
        Also called *Kdes*.
    n_min_bins: :obj:`int` (optional)
        The minimum bin number to be used, usually taken from the window function.
        Also called *bmin*.
    min_segment_length: :obj:`float` (optional)
        The minimum segment length.
        Also called *Lmin*.
    psll: :obj:`float` (optional)
        Peak side-lobe level.
    use_c_core: :obj:`bool`
        Use the C core or the pure Python implementation.


    Returns
    -------
    result: :obj:`openqlab.io.DataFrame`
        calculated frequency data with the columns:
        ps, psd, ps_std, psd_std, enbw, asd

    """
    kwargs["csd"] = True
    lcsd = LCSD(data, *args, **kwargs)
    lcsd.check_inputs()
    return lcsd.run()


def lpsd(data: DataFrame, *args, **kwargs) -> Union[DataFrame, dict]:
    """
    Wrapper for LPSD assuming data in a :obj:`pandas.DataFrame`.
    The sample frequency will be calculated from the index column.

    Example
    -------
    .. code-block:: python

        from lpsd import lpsd
        result = lpsd(data["column"])

    Parameters
    ----------
    data: :obj:`pandas.Series` or :obj:`pandas.DataFrame`
        Assuming time series data and takes one column of a :obj:`pandas.DataFrame`
        or a whole `DataFrame`. If multiple columns are provided, it will calculate
        the spectrum for each column and return a dict of `DataFrame`s with all
        results.
    sample_rate: :obj:`float` (optional)
        Sampling rate of the data. Defaults to calculating the mean difference of the first columns elements.
    window_function: :obj:`Callable` (optional)
        Define a window function, defaults to :obj:`np.kaiser`.
    overlap: :obj:`float` (optional)
        Overlap percentage, usually taken from the window function.
        Defaults to recommended overlap.
    detrending_order: :obj:`int` (optional)
        Order for detrending, 0 = offset, 1 = linear, :obj:`None` to disable.
        Also just called *order*.
    n_frequencies: :obj:`int` (optional)
        The desired number of frequencies.
        Also called *Jdes*.
    n_averages: :obj:`int` (optional)
        The desired number of averages
        Also called *Kdes*.
    n_min_bins: :obj:`int` (optional)
        The minimum bin number to be used, usually taken from the window function.
        Also called *bmin*.
    min_segment_length: :obj:`float` (optional)
        The minimum segment length.
        Also called *Lmin*.
    psll: :obj:`float` (optional)
        Peak side-lobe level.
    use_c_core: :obj:`bool`
        Use the C core or the pure Python implementation.

    Returns
    -------
    result: :obj:`openqlab.io.DataFrame` or :obj:`dict(openqlab.io.DataFrame)`
        calculated frequency data with the columns:
        ps, psd, ps_std, psd_std, enbw, asd

    """

    lcsd = LCSD(data, *args, **kwargs)
    lcsd.check_inputs()
    return lcsd.run()


def lpsd_trad(*args, **kwargs):  # pylint: disable=too-many-arguments
    raise NotImplementedError(
        "Since version 1.0 `lpsd_trad()` is no longer available. Consider using `lpsd()` or "
        "downgrading to version 0.2.2"
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lpsd']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18,<2.0', 'pandas>=1.0,<2.0', 'scipy>=1.5,<2.0']

extras_require = \
{':python_version <= "3.7"': ['importlib-metadata>=1.1.3,<2.0.0']}

setup_kwargs = {
    'name': 'lpsd',
    'version': '1.0.0',
    'description': 'Python and C implementation for logarithmical power spectral density (LPSD) calculation.',
    'long_description': '# Python 3 LPSD algorithm\n\n## Overview\n\nThis repository contains a Python 3 implementation of the LPSD algorithm.\nThe implementation is similiar to the LPSD implementation in the Matlab package LTPDA.\nThe core of the algorithm can be run as Python 3 or as (faster) C code.\nBoth deliver the *same results*.\nTo run the C core the file ltpda_dft.c has to be compiled to a shared library.\n\n\n## Installation\n\nInstall directly with pip:\n```bash\npip install lpsd\n```\n\n\n## Usage\n\nFully working examples can be found in [/doc/examples](https://git.physnet.uni-hamburg.de/gwd/lpsd/-/tree/main/doc/examples).\n\n### With a DataFrame\n\nRecommended interface, direct usage of a `DataFrame`\n\n```python\nimport pandas as pd\nfrom lpsd import lpsd\n# read time series\ndata = pd.read_csv("time_series.csv.gz", index_col=0)\n# select column and calculate\nspectrum = lpsd(data["column"])\n# plot PSD\nspectrum["psd"].plot(logx=True, logy=True)\n```\n\n### Using numpy arrays\n\nUse the traditional method `lpsd_trad`, which uses simple numpy arrays.\n\n\n## References\n\n- [Improved spectrum estimation from digitized time series on a logarithmic frequency axis](https://doi.org/10.1016/j.measurement.2005.10.010)\nAuthors: Michael Tröbs and Gerhard Heinzel\n- [Spectrum and spectral density estimation by the Discrete Fourier transform (DFT), including a comprehensive list of window functions and some new flat-top windows](http://hdl.handle.net/11858/00-001M-0000-0013-557A-5)\nAuthors: Gerhard Heinzel, Albrecht Rüdiger and Roland Schilling\n- [MATLAB Toolbox LTPDA](https://www.elisascience.org/ltpda/)\nAuthor:  Martin Hewitson\n',
    'author': 'Christoph Bode',
    'author_email': 'christoph.bode@aei.mpg.de',
    'maintainer': 'Christian Darsow-Fromm',
    'maintainer_email': 'cdarsowf@physnet.uni-hamburg.de',
    'url': 'https://gitlab.com/uhh-gwd/lpsd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



import math
from pathlib import Path
from sys import platform
from typing import Callable
from warnings import warn

import numpy as np
from scipy.signal import detrend

try:
    import ctypes as ct

    _ctypes_available = True
except ImportError:
    _ctypes_available = False


def c_core_available() -> bool:
    try:
        if not _ctypes_available:
            raise OSError()
        _dft()

        return True
    except OSError as e:
        warn(f"C core not available: {e}", RuntimeWarning)
        return False


def _dft() -> Callable:
    platforms: dict = {
        "win32": dict(libname="ltpda_dft.dll", call="WinDLL"),
        "linux": dict(libname="ltpda_dft.so", call="CDLL"),
        "cygwin": dict(libname="ltpda_dft.so", call="CDLL"),
        "darwin": dict(libname="ltpda_dft.so", call="CDLL"),
    }
    p_dict = platforms[platform]
    dll_abs_path = Path(__file__).parent / p_dict["libname"]
    lib = getattr(ct, p_dict["call"])(str(dll_abs_path), use_last_error=True)

    # function gives pointer to numpy array
    ndpointer = np.ctypeslib.ndpointer
    # make dft function accessible in python
    dft = lib.dft

    # c function to use in python: void dft(double *Pr, double *Vr, long int *Navs, double *xdata, long int nData, long int segLen, double *Cr, double *Ci, double olap, int order)
    # specify return type and argument type of dft function

    dft.restype = None
    dft.argtypes = [
        ct.POINTER(ct.c_double),
        ct.POINTER(ct.c_double),
        ct.POINTER(ct.c_long),
        ndpointer(ct.c_double, flags="C_CONTIGUOUS"),
        ndpointer(ct.c_double, flags="C_CONTIGUOUS"),
        ct.c_long,
        ct.c_long,
        ndpointer(ct.c_double, flags="C_CONTIGUOUS"),
        ndpointer(ct.c_double, flags="C_CONTIGUOUS"),
        ct.c_double,
        ct.c_int,
        ct.c_bool,
    ]
    return dft


def _calc_lcsd_py(
    x1, x2, f, r, m, L, fs, win, psll, order, olap, Lmin, csd
):  # pylint: disable=too-many-arguments,unused-argument,too-many-branches
    """
    Computes the LCSD algorithm like in the Matlab LTPDA implementation,
    generalized to CSD

    Parameters:
        x1 (list of float): first time-series to be processed
        x2 (list of float): second time-series to be processed
        f (list of float): The frequency
        r (list of float): Frequency resolution (not implemented)
        m (int): Bin number
        L (list of float): Segment lengths
        fs (int): Sampling frequency
        psll (float): Peak side-lobe level
        win (function): Window function to be used
        order (int): Order
        olap (float): Overlap (<=1)
        Lmin (int): The minimum segment length
        csd (bool): whether calculating CSD (True) or PSD (False)

    Returns:
        S (list of float): Power spectrum
        Sxx (list of float): Power spectral density
        dev (list of float): Standard deviation PS
        devxx (list of float): Standard deviation PSD
        ENBW (list of float): Equivalent noise bandwidth
    """
    # TODO frequency resolution r not implemented

    # check window function
    win_kaiser = False
    beta = 0
    alpha = 0

    if win is np.kaiser:
        win_kaiser = True
        # calculate kaiser parameter
        alpha = _kaiser_alpha(psll)
        beta = alpha * np.pi

    # define constants
    twopi = 2 * np.pi

    # nx = len(x)
    nf = len(f)
    # initialize outputs
    Sxx = np.zeros(nf, dtype=np.float64)
    S = np.zeros(nf, dtype=np.float64)
    ENBW = np.zeros(nf, dtype=np.float64)
    devxx = np.zeros(nf, dtype=np.float64)
    dev = np.zeros(nf, dtype=np.float64)
    asd = np.zeros(nf, dtype=np.float64)

    # disp_each = _myround(nf / 100) * 10

    x1data = np.array(x1)
    x2data = np.array(x2)

    minReached = False

    for i in range(nf):

        # compute DFT exponent and window
        l = int(L[i])  # segment length

        if not minReached:
            if not win_kaiser:
                window = win(l)
            else:
                # adjust to make window asymmetric (consistent with LTPDA implementation)
                window = win(l + 1, beta)[0:-1]
        if l == Lmin:
            minReached = True

        p = 1j * twopi * m[i] / l * np.arange(0, l)
        C = window * np.exp(p)
        if i == 0:
            C_r = np.array(C.real, dtype=np.float64)
            C_i = np.array(C.imag, dtype=np.float64)
        else:
            C_r = np.append(C_r, C.real)
            C_i = np.append(C_i, C.imag)

        # do segments
        Xr = 0.0
        Qr = 0.0
        Mr = 0.0
        M2 = 0.0

        # Compute the number of averages we want here
        segLen = l  # Segment length
        nData = len(x1data)
        ovfact = 1 / (1 - olap)

        davg = (((nData - segLen)) * ovfact) / segLen + 1
        navg = _myround(davg)

        # Compute steps between segments
        if navg == 1:
            shift = 1.0
        else:
            shift = (float)(nData - segLen) / (float)(navg - 1)
        shift = max(shift, 1.0)

        # changed to 0.0 for python
        start = 0.0

        for j in range(navg):
            # compute start index
            istart = int(_myround(start))
            start = start + shift

            # get segment
            x1s = x1data[istart : istart + l]
            x2s = x2data[istart : istart + l]

            # detrend segment
            _detrend(x1s, order)
            if csd:
                _detrend(x2s, order)
            else:  # x1s is the same data as x2s in this case - simply copy detrending result
                x2s = x1s.copy()

            # make DFT
            a1 = np.dot(C, x1s)
            a2 = np.dot(C, x2s)

            # Welford's algorithm to update mean and variance (see C code)

            if j == 0:
                Mr = a1 * np.conj(a2)
            else:
                Xr = a1 * np.conj(a2)
                Qr = Xr - Mr
                Mr += Qr / j
                M2 += Qr * (Xr - Mr)

        A2ns = 2.0 * Mr.real
        S1 = np.sum(window)
        S12 = S1 * S1
        S2 = np.sum(window**2)

        ENBW[i] = fs * S2 / S12
        Sxx[i] = A2ns / fs / S2
        S[i] = A2ns / S12
        asd[i] = np.sqrt(Sxx[i])

    asdrms, _ = _asdrms(asd, f)

    return [S, Sxx, dev, devxx, ENBW, asd, asdrms]


def _kaiser_alpha(psll):
    """
    KAISER_ALPHA returns the alpha parameter that gives the required input PSLL.

    Taken from C code of Gerhard Heinzel:

    Compute the parameter alpha of Kaiser windows
    from the required PSLL [dB]. Best-fit polynomial
    was obtained from 180 data points between alpha=1
    and alpha=9.95. Maximum error is 0.05
    Maximum error for PSLL > 30 dB is 0.02
    """

    a0 = -0.0821377
    a1 = 4.71469
    a2 = -0.493285
    a3 = 0.0889732

    x = psll / 100
    return ((((a3 * x) + a2) * x) + a1) * x + a0


def _myround(val):
    # implement _myround to be consistent with other implementations
    if (float(val) % 1) >= 0.5:
        x = math.ceil(val)
    else:
        x = round(val)
    return x


def _kaiser_rov(alpha):
    """
    KAISER_ROV returns the recommended overlap for a Kaiser window with parameter alpha.

    Taken from C code of Gerhard Heinzel:

    Compute the 'recommended overlap' (ROV) of Kaiser windows
    from the parameter alpha. Best-fit polynomial
    was obtained from 180 data points between alpha=1
    and alpha=9.95. Maximum error is 1.5%, mainly due
    to insufficient precision in the data points
    """
    a0 = 0.0061076
    a1 = 0.00912223
    a2 = -0.000925946
    a3 = 4.42204e-05
    x = alpha
    return (100 - 1 / (((((a3 * x) + a2) * x) + a1) * x + a0)) / 100


def _ltf_plan(Ndata, fs, olap, bmin, Lmin, Jdes, Kdes):
    """
    Computes the input values needed for the LPSD algorithm

    Parameters:
        Ndata (int): The length of the time-series to be processed
        fs (float): The sample rate of the time-series to be processed
        olap (float): Overlap percentage, usually taken from the window function
        bmin (int): the minimum bin number to be used, usually taken from the window function
        Lmin (float): The minimum segment length
        Jdes (int): The desired number of frequencies
        Kdes (int): The desired number of averages

    Returns:
        f (list of float): The frequency
        r (list of float): Frequency resolution (Hz)
        b (list of float): Bin number
        L (list of float): Segment lengths
        K (list of float): Number of averages
    """

    # set up some variables
    xov = 1 - olap
    fmin = fs / Ndata * bmin
    fmax = fs / 2
    fresmin = fs / Ndata
    freslim = fresmin * (1 + xov * (Kdes - 1))
    logfact = (Ndata / 2) ** (1 / Jdes) - 1

    # prepare outputs
    f = []
    r = []
    m = []
    L = []
    K = []

    # loop over frequencies
    fi = fmin
    while fi < fmax:
        fres = fi * logfact
        # TODO should be less not less equal
        if fres <= freslim:
            fres = np.sqrt(fres * freslim)
        fres = max(fres, fresmin)

        fbin = fi / fres
        if fbin < bmin:
            fbin = bmin
            fres = fi / fbin

        dftlen = _myround(fs / fres)
        dftlen = min(dftlen, Ndata)
        dftlen = max(dftlen, Lmin)

        nseg = _myround((Ndata - dftlen) / (xov * dftlen) + 1)
        if nseg == 1:
            dftlen = Ndata

        fres = fs / dftlen
        fbin = fi / fres

        # store outputs
        f.append(fi)
        r.append(fres)
        m.append(fbin)
        L.append(dftlen)
        K.append(nseg)

        fi = fi + fres

    return [f, r, m, L, K]


def _calc_lcsd(
    x1, x2, f, r, m, L, fs, win, psll, order, olap, Lmin, csd
):  # pylint: disable=too-many-arguments,unused-argument
    """
    Computes the LPSD algorithm like in the Matlab LTPDA implementation

    Parameters:
        x (list of float): The length of the time-series to be processed
        f (list of float): The frequency
        r (list of float): Frequency resolution
        m (int): Bin number
        L (list of float): Segment lengths
        fs (int): Sampling frequency
        psll (float): Peak side-lobe level
        win (function): Window function to be used
        order (int): Order
        olap (float): Overlap percentage, usually taken from the window function
        Lmin (int): The minimum segment length

    Returns:
        S (list of float): Power spectrum
        Sxx (list of float): Power spectral density
        dev (list of float): Standard deviation PS
        devxx (list of float): Standard deviation PSD
        ENBW (list of float): Equivalent noise bandwidth
    """
    olap *= 100

    # check window function
    win_kaiser = False
    beta = 0
    alpha = 0
    if win is np.kaiser:
        win_kaiser = True
        # calculate kaiser parameters
        alpha = _kaiser_alpha(psll)
        beta = alpha * np.pi

    # number of frequency bins
    nf = len(f)

    # get C core
    dft = _dft()

    # initialize outputs
    Sxx = np.zeros(nf, dtype=np.float64)
    S = np.zeros(nf, dtype=np.float64)
    ENBW = np.zeros(nf, dtype=np.float64)
    devxx = np.zeros(nf, dtype=np.float64)
    dev = np.zeros(nf, dtype=np.float64)
    asd = np.zeros(nf, dtype=np.float64)

    # disp_each = _myround(nf / 100) * 10
    min_reached = False

    # initialize dft outputs
    Pr = ct.c_double(0)
    Vr = ct.c_double(0)
    nsegs = ct.c_long(0)

    # pointer to data
    x1data = np.array(x1)
    # length should be the same for both inputs
    nData = ct.c_long(len(x1))
    x2data = np.array(x2)

    for i in range(nf):

        # compute DFT exponent and window
        l = int(L[i])  # segment length

        if not min_reached:
            if not win_kaiser:
                window = win(l)
            else:
                # adjust to make window asymmetric (consistent with LTPDA implementation)
                window = win(l + 1, beta)[0:-1]
        if l == Lmin:
            min_reached = True

        p = 1j * 2 * np.pi * m[i] / l * np.arange(0, l)
        C = window * np.exp(p)

        Cr = np.array(C.real, dtype=np.float64)  # Real part of DFT coefficients
        Ci = np.array(C.imag, dtype=np.float64)  # Imag part of DFT coefficients

        # Core DFT part implemented in C file
        dft(
            ct.byref(Pr),
            ct.byref(Vr),
            ct.byref(nsegs),
            x1data,
            x2data,
            nData,
            ct.c_long(l),
            Cr,
            Ci,
            ct.c_double(olap),
            ct.c_int(order),
            ct.c_bool(csd),
        )

        A2ns = 2.0 * Pr.value
        B2ns = 4.0 * Vr.value / nsegs.value
        S1 = sum(window)
        S12 = S1 * S1
        S2 = sum(window**2)
        ENBW[i] = fs * S2 / S12
        # Scale PS/PSD
        Sxx[i] = A2ns / fs / S2
        S[i] = A2ns / S12
        # Scale sqrt(variance)
        devxx[i] = np.sqrt(B2ns / fs**2 / S2**2)
        dev[i] = np.sqrt(B2ns / S12**2)
        asd[i] = np.sqrt(Sxx[i])

    asdrms, _ = _asdrms(asd, f)

    return [S, Sxx, dev, devxx, ENBW, asd, asdrms]


def _asdrms(asd_in, freq_in, f_start=None):  # TODO: allow user to specify f_start here
    """
    Calculates the high-to-low RMS of an ASD.

    Conor Mow-Lowry June 20 2016, updated November 13 2017 (original MATLAB code)
    Artem Basalaev 20 April 2022 (python version)

    Parameters
    ----------
     asd_in: array_like
        input amplitude spectra density, from low-to-high frequency
     freq_in: array_like
        linearly-spaced frequency vector for ASD [Hz]
     f_start: float
        frequency to begin accumulating RMS (optional)

    Returns
    -------
    rms_out: numpy::Array
        high-to-low frequency cumulative RMS
    freq_out: numpy::Array
        (cropped) output frequency vector (optional)
    """

    freq_out = freq_in
    if f_start is not None:  # Cutting away data above f_start
        n_cut = np.where(freq_in > f_start)[0][0]  # Index for start frequency
        freq_out = freq_in[0:n_cut]
        asd_in = asd_in[0:n_cut]

    bin_widths = np.diff(freq_out)
    bin_widths = np.append(
        bin_widths[0], bin_widths
    )  # add an extra "bin" of the same size as first one
    rms_out = np.sqrt(np.flip(np.cumsum(np.flip(asd_in * asd_in * bin_widths))))
    return rms_out, freq_out


def _detrend(data, order):
    """
    Detrend time series segment. Overwrites input `data` with detrended version

    Parameters
    ----------
     data: array_like
        input time series segment
     order: float
        detrending order
    """

    if order == -1:
        return  # do nothing
    elif order == 0:
        data = data - np.mean(data)
    elif order == 1:
        detrend(data, overwrite_data=True)
    else:
        # TODO implement scipy detrending
        # data = polydetrend(data, order)
        warn(
            "Polynomial detrending is not implemented in Python, yet. Try the C version.",
            UserWarning,
        )
        return
